#include <iostream>
#include <fstream>
#include <strstream>
#include <sstream>
#include <iomanip>
#include <map>
#include <list>
#include <string>
#include <locale>
#include <vorbis/vorbisfile.h>
#include <boost/scoped_array.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/tokenizer.hpp>

#include <windows.h>

#include <atlbase.h>

namespace fs = boost::filesystem;

typedef signed __int8 int8_t;
typedef signed __int16 int16_t;
typedef signed __int32 int32_t;
typedef unsigned __int8 uint8_t;
typedef unsigned __int16 uint16_t;
typedef unsigned __int32 uint32_t;

// ZkcCX^@^NX
class RNG_MT
{
private:
  enum {
    N = 624,
    M = 397,
    MATRIX_A = 0x9908b0dfUL,
    UPPER_MASK = 0x80000000UL,
    LOWER_MASK = 0x7FFFFFFFUL
  };

private:
  uint32_t mt[N];
  int32_t mti;

public:
  explicit RNG_MT(uint32_t s)
  {
    init(s);
  }

  void init(uint32_t s)
  {
    mt[0] = s;
    for(mti = 1; mti < N; ++mti) {
      mt[mti] =
        (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti);
    }
  }

  uint32_t next_int32()
  {
    uint32_t y;
    static const uint32_t mag01[2]={0x0UL, MATRIX_A};
    if (mti >= N) {
      int kk;
      if (mti == N+1)
          init(5489UL);

      for (kk=0;kk<N-M;kk++) {
        y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
        mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
      }
      for (;kk<N-1;kk++) {
        y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
        mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
      }
      y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
      mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

      mti = 0;
    }
  
    y = mt[mti++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
  }
};

// zVA[JCu̓NX
class HisotenArchive
{
	friend class Entry;
private:
	struct HisotenEntryInfo {
		uint32_t offset;
		uint32_t size;
		std::string name;
	};
	typedef std::list<HisotenEntryInfo> EntryList;

public:
	class Entry
	{
		friend class HisotenArchive;
	private:
		HisotenArchive::EntryList::iterator it;
		HisotenArchive *parent;
	private:
		Entry(HisotenArchive &p, HisotenArchive::EntryList::iterator vi) :
			parent(&p), it(vi) {}

	public:
		Entry() : parent(NULL), it() {}
		Entry(const Entry &orig) : parent(orig.parent), it(orig.it) {}

	public:
		uint32_t GetSize()
		{
			return it->size;
		}

		const char * GetName()
		{
			return it->name.c_str();
		}

		bool Extract(std::ostream &os)
		{
			return parent->DoExtract(it, os);
		}

		operator bool()
		{
			return parent != NULL;
		}
	};

public:
	bool Open(const char *name)
	{
		is.open(name, std::ios::binary);
		if(is.fail()) return false;

		std::ifstream::pos_type begpos = is.tellg();
		is.seekg(0, std::ios_base::end);
		std::ifstream::pos_type endpos = is.tellg();
		is.seekg(0, std::ios_base::beg);
		std::ifstream::pos_type file_size = endpos - begpos;
		if(file_size < 6) return false;

		uint16_t list_count;
		is.read(reinterpret_cast<char *>(&list_count), sizeof(list_count));
		if(is.eof() || is.fail() || is.gcount() != sizeof(list_count)) return false;
		uint32_t list_size;
		is.read(reinterpret_cast<char *>(&list_size), sizeof(list_size));
		if(is.eof() || is.fail() || is.gcount() != sizeof(list_size)) return false;

		boost::scoped_array<char> list_buf(new char[list_size]);
		is.read(list_buf.get(), list_size);
		if(is.eof() || is.fail() || is.gcount() != list_size) return false;

		mode = 0;
		RNG_MT mt(list_size + 6);
		for(uint32_t i = 0; i < list_size; ++i)
			list_buf[i] ^= mt.next_int32() & 0xFF;
		if(ReadList(list_buf.get(), list_count, list_size, file_size))
			return true;

		mode = 1;
		entry_list.clear();
		uint8_t k = 0xC5, t = 0x83;
		for(uint32_t i = 0; i < list_size; ++i) {
		  list_buf[i] ^= k; k += t; t +=0x53;
		}
		return ReadList(list_buf.get(), list_count, list_size, file_size);
	}

	void Close()
	{
		is.close();
		entry_list.clear();
	}

	bool EntryFirst()
	{
		it = entry_list.begin();
		return it != entry_list.end();
	}

	bool EntryNext()
	{
		if(it == entry_list.end()) return false;
		return ++it != entry_list.end();
	}

	uint32_t GetSize()
	{
		return it->size;
	}

	const char * GetName()
	{
		return it->name.c_str();
	}

	bool Extract(std::ostream &os)
	{
        return DoExtract(it, os);
	}

	class Entry GetEntry()
	{
		return Entry(*this, it);
	}

	int GetMode()
	{
		return mode;
	}

private:
	bool ReadList(char *list_buf, uint32_t list_count, uint32_t list_size, uint32_t file_size)
	{
		const char *p = list_buf;
		uint32_t offset = 0;
		for(uint16_t i = 0; i < list_count; ++i) {
			if(offset + 9 > file_size) return false;
			uint8_t name_len = *reinterpret_cast<const uint8_t *>(p + 8);
			if(offset + 9 + name_len > file_size) return false;
			const HisotenEntryInfo e = {
				*reinterpret_cast<const uint32_t *>(p),
				*reinterpret_cast<const uint32_t *>(p + 4),
				std::string(p + 9, p + 9 + name_len)
			};
			if(e.offset < list_size + 6 || e.offset > file_size) return false;
			if(e.size > file_size - e.offset) return false;
			if(e.name == "") return false;
			entry_list.push_back(e);
			p += 9 + name_len;
			offset += 9 + name_len;
		}
		return true;
	}

private:
	bool DoExtract(EntryList::iterator entry, std::ostream &os)
	{
		boost::scoped_array<char> data(new char[entry->size]);

        is.seekg(entry->offset, std::ios::beg);
        is.read(reinterpret_cast<char *>(data.get()), entry->size);
		if(is.eof() || is.fail() || is.gcount() != entry->size) return false;

        uint8_t k = (entry->offset >> 1) | 0x23;
        for(uint32_t i = 0; i < entry->size; ++i)
          data[i] ^= k;

        os.write(data.get(), entry->size);
        return true;
	}

private:
	int mode;
	EntryList entry_list;
	EntryList::iterator it;

	std::ifstream is;
};

// Ogg Vorbis ǍNX
class ovfile {
private:
	bool opened;
	OggVorbis_File vf;

private:
	static size_t stream_read(void *ptr, size_t size, size_t nmemb, void *datasource)
	{
		std::iostream *s = reinterpret_cast<std::iostream *>(datasource);
		s->read(static_cast<char *>(ptr), size * nmemb);
		return s->gcount() / size;
	}

	static int stream_seek(void *datasource, ogg_int64_t offset, int whence)
	{
		std::iostream *s = reinterpret_cast<std::iostream *>(datasource);
		std::iostream::pos_type type;
		switch(whence) {
		case SEEK_SET: type = std::ios::beg; break;
		case SEEK_CUR: type = std::ios::cur; break;
		case SEEK_END: type = std::ios::end; break;
		}
		s->clear();
		s->seekg(offset, type);
		return s->eof() || s->fail() || s->bad() ? -1 : 0;
	}

	static int stream_close(void *datasource)
	{
		return 0;
	}

	static long stream_tell(void *datasource)
	{
		std::iostream *s = reinterpret_cast<std::iostream *>(datasource);
		return s->tellg();
	}

	static ov_callbacks stream_callbacks;

public:
	ovfile() : opened(false) {}

	~ovfile() {
		if(opened) ov_clear(&vf);
	}

	bool open_stream(std::iostream &s, char *initial, long ibytes) {
		s.clear();
		s.seekg(0, std::ios::beg);
		if(ov_open_callbacks(reinterpret_cast<void *>(&s), &vf, initial, ibytes, stream_callbacks) < 0)
			return false;

		opened = true;
		return true;
	}
	vorbis_info * info() {
		if(!opened) return NULL;
		return ov_info(&vf, -1);
	}

	ogg_int64_t pcm_total() {
		if(!opened) return NULL;
		return ov_pcm_total(&vf, -1);
	}

	long read(char *buf, long len) {
		if(!opened) return NULL;

		int bitstream;
		long total = 0;
		long readlen;

		while(len > total) {
			readlen = ov_read(
				&vf,
				buf, len - total,
				0, 16 >> 3, 1, &bitstream);
			if (readlen == OV_HOLE ||
				readlen == OV_EBADLINK ||
				readlen == 0) break;
			total += readlen;
			buf   += readlen;
		}

		return total;
	}
};

ov_callbacks ovfile::stream_callbacks = {
	ovfile::stream_read,
	ovfile::stream_seek,
	ovfile::stream_close,
	ovfile::stream_tell
};

typedef std::map<std::string, HisotenArchive::Entry> entry_map;
typedef std::list<std::pair<std::string, std::string> > title_list;

// A[JCugq̃Gg𒊏o
void FindArchive(HisotenArchive &ar, const wchar_t *extension, entry_map &ret)
{
	if(ar.EntryFirst()) {
		do {
			fs::wpath name(ATL::CA2W(ar.GetName()));
			if(fs::extension(name) == extension) {
				std::string basename =
					ATL::CW2A(fs::basename(name).c_str());
				ret[basename] = ar.GetEntry();
			}
		} while(ar.EntryNext());
	}
}

// sflf[^烊[W𒊏o
bool ConvertSFLToRegion(std::strstream &sfl, uint32_t size, std::pair<uint32_t, uint32_t> &ret)
{
	MMIOINFO info;
	memset(&info, 0, sizeof(info));
	info.pchBuffer = sfl.str();
	info.fccIOProc = FOURCC_MEM;
	info.cchBuffer = size;

	struct HMMIOHandle {
	private:
		HMMIO mmio;
	public:
		HMMIOHandle(HMMIO vi) : mmio(vi) {}
		~HMMIOHandle() { if(mmio) mmioClose(mmio, 0); }
		operator HMMIO&() { return mmio; }
	} mmio = mmioOpen(NULL, &info, MMIO_READ);
	if(!mmio) return false;

	MMCKINFO cksfpl;
	cksfpl.fccType = mmioFOURCC('S', 'F', 'P', 'L');
	if(mmioDescend(mmio, &cksfpl, NULL, MMIO_FINDRIFF) != MMSYSERR_NOERROR)
		return false;

	MMCKINFO ckcue;
	ckcue.ckid = mmioFOURCC('c', 'u', 'e', ' ');
	if(mmioDescend(mmio, &ckcue, &cksfpl, MMIO_FINDCHUNK) != MMSYSERR_NOERROR)
		return false;
	if(ckcue.cksize != 0x1c)
		return false;

	char cue[0x1c];
	if(mmioRead(mmio, cue, 0x1c) < 1)
		return false;

	MMCKINFO ckadtl;
	ckadtl.fccType = mmioFOURCC('a', 'd', 't', 'l');
	if(mmioDescend(mmio, &ckadtl, NULL, MMIO_FINDLIST) != MMSYSERR_NOERROR)
		return false;

	MMCKINFO ckltxt;
	ckltxt.ckid = mmioFOURCC('l', 't', 'x', 't');
	if(mmioDescend(mmio, &ckltxt, &ckadtl, MMIO_FINDCHUNK) != MMSYSERR_NOERROR)
		return false;
	if(ckltxt.cksize != 0x14)
		return false;

	char ltxt[0x14];
	if(mmioRead(mmio, ltxt, 0x14) < 1)
		return false;

	ret.first = *reinterpret_cast<uint32_t *>(cue + 8) * 4;
	ret.second = *reinterpret_cast<uint32_t *>(ltxt + 4) * 4;
	return true;
}

// music.cv1y^CgXgɕϊ
bool ConvertMusicCV1ToTitleMap(HisotenArchive &ar, title_list &titles)
{
	// music.cv1 T
	HisotenArchive::Entry entry;
	if(ar.EntryFirst()) {
		do {
			if(!strcmp("data/csv/system/music.cv1", ar.GetName())) {
				entry = ar.GetEntry();
				break;
			}
		} while(ar.EntryNext());
	}
	if(!entry) return false;

	// o
	std::strstream buf;
	if(!entry.Extract(buf)) return false;

	// cv0/csvfR[h
	uint8_t k, t, s;
	if(ar.GetMode() == 0) {	// ̌
		k = 0x66; t = 0x65; s = 0x4F;
	} else {				// łȂ
		k = 0x8B; t = 0x71; s = 0x95;
	}
	char *p = buf.str();
	uint32_t size = entry.GetSize();
	for(uint32_t i = 0; i < size; ++i) {
		p[i] ^= k;
		k += t; t += s;
	}

	// csv^Cg}bv𐶐
	typedef boost::escaped_list_separator<char> separator;
	typedef boost::tokenizer<separator> tokenizer;
	separator sep;

	std::string line;
	std::getline(buf, line);
	for(;;) {
		std::getline(buf, line);
		if(buf.eof()) break;
		if(line.size() > 0 && *line.rbegin() == '\r')
			line.erase(--line.end());
		std::vector<std::string> el;	
		tokenizer tokens(line, sep);
		std::copy(tokens.begin(), tokens.end(), std::back_inserter(el));
		if(el.size() != 2) {
			titles.clear();
			return false;
		}

		titles.push_back(std::make_pair(fs::basename(el[0]), el[1]));
	}
	return true;
}

// Ogg Vorbisf[^PCMf[^ɕϊ
bool ConvertOggToPCM(std::iostream &ogg, std::ostream &os, uint32_t &write_bytes)
{
	ovfile vf;
	if(!vf.open_stream(ogg, NULL, 0))
		return false;
	vorbis_info *vi = vf.info();
	if(vi == NULL)
		return false;

	uint32_t buf_bytes = vf.pcm_total() * vi->channels * 2;
	boost::scoped_array<char> buf
		(new char[buf_bytes]);
	
	if(vf.read(buf.get(), buf_bytes) != buf_bytes)
		return false;
	os.write(buf.get(), buf_bytes);
	if(os.fail() || os.bad())
		return false;

	write_bytes = buf_bytes;
	return true;
}

int main(int argc, char *argv[])
{
	std::locale loc = std::locale();
	loc = std::locale(loc, "japanese", std::locale::ctype);
	std::locale::global(loc);

	if(argc < 2) {
		std::cout
			<< "gF" << std::endl
			<< "zVtH_hbOhbv邩AR\[" << std::endl
			<< "gen105dat [zVtH_̃pX]Ƒł܂B" << std::endl << std::endl;
		system("pause");
		return 0;
	}

	// pX̐`FbNdatpX
	std::string apath, bpath;
	fs::wpath indir;
	try {
		indir = ATL::CA2W(argv[1]);
		if(!fs::is_directory(indir)) {
			std::cerr << "fBNgw肵ĂB" << std::endl;
			return 1;
		}
		apath = ATL::CW2A((indir / L"th105a.dat").external_file_string().c_str());
		bpath = ATL::CW2A((indir / L"th105b.dat").external_file_string().c_str());
	} catch(std::exception &e) {
		std::cerr << e.what() << std::endl;
		system("pause");
		return 1;
	}
	
	// th105a.dat J
	HisotenArchive ar;
	if(!ar.Open(apath.c_str())) {
		std::cerr << "th105a.datJ܂" << std::endl;
		system("pause");
		return 1;
	}

	// music.cv1𒊏o
	title_list titles;
	ConvertMusicCV1ToTitleMap(ar, titles);

	// th105b.dat J
	ar.Close();
	if(!ar.Open(bpath.c_str())) {
		std::cerr << "th105b.datJ܂" << std::endl;
		system("pause");
		return 1;
	}

	// gqogg, sfl̃Gg𒊏o
	entry_map ogg_entry;
	entry_map sfl_entry;
	FindArchive(ar, L".ogg", ogg_entry);
	FindArchive(ar, L".sfl", sfl_entry);
	if(ogg_entry.empty()) {
		std::cerr << "yf[^݂܂" << std::endl;
		system("pause");
		return 1;
	}
	// o̓t@CJ
	fs::ofstream os(indir / L"music.dat", std::ios::binary);
	if(os.fail()) {
		std::cerr << "music.dat쐬ł܂" << std::endl;
		system("pause");
		return 1;
	}
	fs::ofstream txtos(indir / L"titles_th105.txt");
	if(txtos.fail()) {
		std::cerr << "titles_th105.txt쐬ł܂" << std::endl;
		system("pause");
		return 1;
	}

	// ݒwb_o͂
	txtos << "@th105\\music.dat,zV@` Scarlet Weather Rhapsody" << std::endl;

	// oggt@C̃GgȂ߂
	for(title_list::iterator it = titles.begin(); it != titles.end(); ++it) {
		std::iostream::pos_type base = os.tellp();

		// oggt@CfR[hyяo͂
		entry_map::iterator e = ogg_entry.find(it->first);
		if(e != ogg_entry.end()) {
			std::strstream ogg;
			uint32_t write_bytes;
			if(e->second.Extract(ogg) && 
			   ConvertOggToPCM(ogg, os, write_bytes)) {
				// [Wsflf[^狁߂
				// ΑŜ[Wɂ
				std::pair<uint32_t, uint32_t> region(0, write_bytes);
				entry_map::iterator sflit = sfl_entry.find(it->first);
				if(sflit != sfl_entry.end()) {
					std::strstream sfl;
					if(sflit->second.Extract(sfl)) {
						ConvertSFLToRegion(sfl, sflit->second.GetSize(), region);
					}
				}
				// 1Gg̐ݒo͂
				txtos << std::hex << std::setw(8) << std::setfill('0') << base << ","
					  << std::hex << std::setw(8) << std::setfill('0') << region.first << ","
					  << std::hex << std::setw(8) << std::setfill('0') << region.second << ","
					  << it->second << std::endl;
			}
		}
	}

	return 0;
}
